var doBuild = true; // true = build the program
                    // false = write a build script
var quiet = false;  
var verbose = false; // not a command line option
var stopEarly = false;
                    
function objMatches(a,platform,mode)
{
  var ret = ((!a.get('platform') || a.get('platform') == (platform?platform.get('name'):'')) && 
          (!a.get('mode') || a.get('mode') == (mode?mode.get('name'):'')));
   return ret;
}

function getProperty(name,node,platform,mode,delimiter)
{
   var d = '';
   if (delimiter == null) delimiter = '';
   var l = node.find(name);
   for (var i in l)
   { 
       if (objMatches(l[i],platform,mode))
         d += l[i].cdata+delimiter ;
   }
   return d;
}

function getAttribute(name,node,delimiter)
{
   var l = node.get(name);
   if (delimiter == null) delimiter = '';
   if (l) return l+delimiter;
   return '';
}
     
//BCC Error E2285 rs/tbl_main.cpp 78: Message
//gcc filename:error:line: description
function scanErrors(msgs,errors)
{
 var l;
 while(l=msgs.readln())
 {  
  if (l.indexOf("Error") == 0) //Borland C++
  {
   return false;
  }
  if (l.search(/\w+\:\d+\:/) >= 0 && l.indexOf("warning") == -1) //gcc
  {
   return false;
  }
  if (l.indexOf("(.text") >= 0) //gcc link
  {
   return false;
  }
 }
 return true;
}

function runCommand(cmd,errors)
{
   if (verbose) writeln(cmd);
   if (cmd.substr(0,2) == '#!') return true;
   var w = cmd.match(/[^ "]+|"[^"]*"/g);
   for (var x =0; x<w.length; x++) w[x] = w[x].replace(/^"|"$/g,'')
   if (w[0] == 'cp' || w[0] == 'copy')
   {
    system.copy(w[1],w[2])
    return true;
   }
   if (w[0] == 'mv' || w[0] == 'move')
   {
    system.move(w[1],w[2],true)
    return true;
   }
   if (w[0] == 'echo')
   {
    writeln(cmd.replace(/^echo/,''));
    return true;
   }
   var msgs = new Stream;
   var out 
   if (cmd.indexOf('>') != -1)
   {
     cmd = cmd.split('>')
     out = new Stream(cmd[1].replace(/^\s+/,''),'wt')
     cmd = cmd[0].replace(/\s+$/,'')
   }
   var pc = new Stream('exec://'+cmd);  
   var se = pc.stderr;
   while (pc.canWrite)
     {
      sleep(10);
      while (pc.canRead)    
      {
       var c = pc.read(1);
       if (verbose) write(c);
       if (out) out.write(c)
       else msgs.write(c);
      }
      while (se.canRead)
      {
       var c = se.read(1);
       if (verbose) write(c);
       msgs.write(c);
      }
     }
   pc.close();
   if (out) out.close()
   msgs.rewind();
   if (!scanErrors(msgs))   
    {
     msgs.rewind();
     errors.writeln(cmd);
     errors.append(msgs);
     return false;
    }
   return true;
}

function rejoin(template,array)
{
 var s = '';
 for (var a in array) if (array[a]) s += template.replace(/\$value/g,array[a]);
 return s;
}

function compileFile(cmds,compiler,target,platform,mode,group,file)
{
  var define =  getProperty('define',mode,platform,null,';');
      define += getProperty('define',platform,null,mode,';');
      define += getProperty('define',target,platform,mode,';');
      define += getProperty('define',group,platform,mode,';');
      define += getProperty('define',file,platform,mode,';');
      define += getProperty('define',compiler,platform,mode,';');

  var include = group.get('path');
      if (include) include += ';';
      include +=  getAttribute('include',mode,';');
      include += getAttribute('include',platform,';');
      include += getAttribute('include',group,';');
      include += getAttribute('include',file,';');
         
  var cmd = compiler.get('path');
  var params = compiler.get('parameters');
  var infile = group.get('path') + file.cdata;  
  var outfile = file.cdata.replace(/\.\w*$/,compiler.get('output'));
  
  var options = getProperty('options',compiler,platform,mode,' ');
  
  var templates = compiler.find('template','name','define');
  if (templates.length)
  {
    define = rejoin(templates[0].cdata,define.split(';'))
  }
  
  templates = compiler.find('template','name','include');
  if (templates.length)
  {
    include = rejoin(templates[0].cdata,include.split(';'))
  }
  
  params = params.replace(/\$input/g,infile);
  params = params.replace(/\$output/g,outfile);
  params = params.replace(/\$options/g,options);
  params = params.replace(/\$define/g,define);
  params = params.replace(/\$include/g,include);
  params = params.replace(/\$intermediate/g,target.get('intermediate'));
  
  if (doBuild)
  { 
    writeln(file.cdata);
    if (!runCommand(cmd + ' ' + params,cmds)) 
       throw(file.cdata); //+'\r\n'+cmd+' '+params);
  }
  else
  {
   cmds.writeln(cmd + ' ' + params);  
  }
  return outfile;
}

function doConfigure(commands,configure,platform,mode)
{
 try 
 {
  for (var i in configure)
  {
    if (objMatches(configure[i],platform,mode)) 
    {
      var cmd = configure[i].cdata;
      if (doBuild)
      { 
        writeln(cmd);
        if (!runCommand(cmd,commands)) 
           throw(cmd); //+'\r\n'+cmd+' '+params);
      }
      else
      {
        commands.writeln(cmd);  
      }    
    }
  }
 } catch(err)
 {
   writeln(err)
   return false
 }
}
//defines and includes are additive options

function doCompile(cmds,project,target,platform,mode)
{
  var defaultCompiler = platform.find('compile')[0];
  
  var objFiles = [];
  
  var gnames = target.find('group');
  var groups = [target]; //include top-level <file> tags
  for (var i in gnames)
   groups = groups.concat(project.find('group','name',gnames[i].cdata));
  
  for (var i in groups)
  {
    if (objMatches(groups[i],platform,mode))
    {
      var files = groups[i].find('file');
      for (var j in files)
      {
        if (objMatches(files[j],platform,mode))
        {
          var filename = files[j].cdata;
          var type = filename.substr(filename.lastIndexOf('.'));
          var compiler = platform.find('compile','input',type)[0];
          if (!compiler) compiler = defaultCompiler;
          objFiles.push(compileFile(cmds,compiler,target,platform,mode,groups[i],files[j]));
        }             
      }
     }
  } 
  return objFiles;
}

function findFile(target,linker,file)
{
 if (!linker.get('resolve')) return file;
 return target.get('intermediate')+file;
 return file;
}

function doLink(cmds,objFiles,target,platform,mode)
{
  //link step

  var linker = platform.find('link','mode',mode)[0];
  if (!linker) linker = platform.find('link')[0];  
  var cmd = linker.get('path');
  var params = linker.get('parameters');
  var infiles = ''; 
  var rfiles = '';
  for (var i in objFiles) 
   { 
     if (objFiles[i].indexOf('.o') > 0) infiles += findFile(target,linker,objFiles[i])+' ';
     else rfiles += findFile(target,linker,objFiles[i]) + ' ';
    }
  var options = getProperty('options',linker,platform,mode,' ');
  
  //<link parameters="$options, C0X32 $input,$output,$mapfile,IMPORT32 CW32MT,$deffile " >
  //<target platform=win32 output="jsdb.exe" mapfile="jsdb.map" deffile="jsdb.def" intermediate="obj\">

  var libraries = target.get('libraries');

  var templates = linker.find('template','name','library');
  if (templates.length)
  {
      libraries = rejoin(templates[0].cdata,libraries.split(';'))
  }
  params = params.replace(/\$input/g,infiles);
  params = params.replace(/\$resource/g,rfiles);
  params = params.replace(/\$output/g,target.get('output'));
  params = params.replace(/\$mapfile/g,target.get('mapfile'));
  params = params.replace(/\$deffile/g,target.get('deffile'));
  params = params.replace(/\$libraries/g,libraries);
  params = params.replace(/\$options/g,options);  
  params = params.replace(/\$intermediate/g,target.get('intermediate'));
  
  if (doBuild)
  { 
    writeln(target.get('output'));
    if (!runCommand(cmd + ' ' + params,cmds)) 
      throw(target.get('output')+'\r\n'+cmd+' '+params);
  }
  else
  {
    cmds.writeln(cmd + ' ' + params); 
  }
}

function runMake(project,platform)
{ 
 var targetName = null;
 var mode = null; 
 
 while (system.arguments.length)
 { 
   var p = system.arguments.shift()
   if (p.toLowerCase() == '/preview')
     doBuild = false;
   else if (p.toLowerCase() == '/quiet')
     quiet = true;
   else if (p.toLowerCase() == '/verbose')
     verbose = true;
   else if (p.toLowerCase() == '/stopwarn')
     stopEarly = true;
   else if (p[0] == '/') 
   {
    var x = project.find('mode','name',p.substr(1));
    if (!mode && x.length)
      mode = x[0];
   }
   else 
      targetName = p;
   // else other flags     
 }

 if (!mode) mode = project.find('mode')[0];
 
 var targets = project.find('target');

 if (doBuild) writeln('Building ',project.get('name'),' for ',platform.get('name'));
 
 var commands = new Stream; 
 
 // run platform-dependent configure scripts
 
 doConfigure(commands,platform.find('configure'),null,mode);
 doConfigure(commands,project.find('configure'),platform,mode);
 
 // look over potential targets
 try {
 for (var i in targets)
  {
   if (targetName && targetName != targets[i].get('name')) continue;

   if (objMatches(targets[i],platform,mode)) 
   {
     if (doBuild) 
     { 
       writeln('Target: ',targets[i].get('output'))
       writeln('Platform: ',platform.get('name'))
       writeln('Mode: ',mode.get('name'))
     }
     doConfigure(commands,targets[i].find('configure'),platform,mode);
     var objs = doCompile(commands,project,targets[i],platform,mode);
     doLink(commands,objs,targets[i],platform,mode);
   }
  } 
 } catch(err) {writeln("Error in ", err); }

 // run platform-dependent configure scripts
 
 doConfigure(commands,platform.find('finish'),null,mode);
 doConfigure(commands,project.find('finish'),platform,mode);
 
 // done!
 if (!quiet)
 {
   commands.rewind();
   write(commands);
 }
 commands.rewind();
 log = new Stream('make.log','wt');
 log.append(commands);
}

if (typeof(system.arguments) == 'undefined' || system.arguments.length == 0)
{
  writeln("Usage: jsdb make.js jsdb.project win32 [target] [/debug] [/preview] [/quiet] [/verbose] [/stopwarn]") 
}
else
{
 load('xml.js')

 var project = XML.read(new Stream(system.arguments.shift()),'target,project,mode,configure,finish,define,compile,link,group,file,platform,options,template');

 if (system.arguments.length)
 {
   var platformName = system.arguments.shift()
   var platform = project.find('platform','name',platformName)[0];
   if (platform) 
     runMake(project,platform)
   else
     writeln("Unknown platform: ",platform)
 }
 else
 {
  writeln("Platforms")
  var l = project.find('platform');
  for (var i in l) writeln(" ",l[i].get('name'));
  
  writeln("Modes")
  var l = project.find('mode');
  for (var i in l) writeln(" ",l[i].get('name'));
  
  writeln("Targets")
  var l = project.find('target');
  for (var i in l) writeln(" ",l[i].get('name'),": ",l[i].get('output')," (",l[i].get('platform'),")");
 }
}
