#pragma hdrfile "jsdb.csm"
#include "rslib.h"
#include "jsdb.h"
#include "rs/wrap_jsdb.h"
#pragma hdrstop

#include "sqlite/sqlite3.h"

class SQLite
{
    public:
    sqlite3* db;
    TStr lastError;
    SQLite(const char* name = NULL) : db(0)
    {  // Database filename (UTF-8)
       int r = sqlite3_open_v2( name ? name : ":memory:",
          &db, SQLITE_OPEN_READWRITE| SQLITE_OPEN_CREATE,NULL);
       if (r != SQLITE_OK)
       {
        sqlite3_close(db);
        db = 0;
        lastError = itos(r);
       }
    }
    ~SQLite()
    {
     if (db) sqlite3_close(db);
    }
};

struct CBContext
{
  JSContext* cx;
  int argc;
  jsval* argv;
};

static int callback(void *NotUsed, int argc, char **argv, char **azColName)
{
    CBContext* c = (CBContext*)NotUsed;
  int i;
  for(i=0; i<argc; i++){
    printf("%s = %s\n", azColName[i], argv[i] ? argv[i] : "NULL");
  }
  printf("\n");
  return 0;
}

WRAP(SQLite,Open)
{
  sqlite3 *db;
  char *zErrMsg = 0;
  int rc;

  if( argc!=3 ){
    fprintf(stderr, "Usage: %s DATABASE SQL-STATEMENT\n", argv[0]);
    exit(1);
  }
  rc = sqlite3_open(argv[1], &db);
  if( rc ){
    fprintf(stderr, "Can't open database: %s\n", sqlite3_errmsg(db));
    sqlite3_close(db);
    exit(1);
  }
  rc = sqlite3_exec(db, argv[2], callback, 0, &zErrMsg);
  if( rc!=SQLITE_OK ){
    fprintf(stderr, "SQL error: %s\n", zErrMsg);
    sqlite3_free(zErrMsg);
  }
  sqlite3_close(db);
  return 0;
}

WRAP(SQLite,Close)
{
 CLOSEPRIVATE(SQLite);
 RETBOOL(true);
}

void SQLite_JSFinalize(JSContext *cx, JSObject *obj)
{
// if (JS_GetParent(cx,obj)) return;
 DELPRIVATE(SQLite);
// JS_SetPrivate(cx,obj,NULL);
}

JSBool
SQLite_SQLite(JSContext *cx, JSObject *obj, uintN argc, jsval *argv,
    jsval *rval)
{
 GETENV;
 if (!Env) return JS_FALSE;

 SQLite * dt = NULL;

 try {
 if (argc == 1)
 {
     if (ISSTR(0))
     {
        GETUTF8(0);
        dt = new SQLite(s0);
     }
     else
        ERR_TYPE(SQLite,new,1,String);
 }
 else
     dt = new SQLite();
  dt = OpenSQLite(STR(0),Env->SQLiteEnv);
 else if (argc == 0)
  dt = new SpreadsheetSQLite;
 } catch (...) {dt = NULL;}

 if (!dt)
 {
  if (Env->errorOnFailure)
    return JS_FALSE;
  ERR_MSG(SQLite,"Open SQLite failed","");
 }

 if (dt)
 {
  SETPRIVATE(obj,SQLite,dt,true,NULL);
 }
// JS_SetPrivate(cx,obj,dt);

 return JS_TRUE;
}

static JSPropertySpec SQLite_properties[] = {
    {"name",      0,   JSPROP_ENUMERATE|JSPROP_READONLY|JSPROP_PERMANENT,SQLite_JSGet},
    {"rowCount",  1,   JSPROP_ENUMERATE|JSPROP_READONLY|JSPROP_PERMANENT,SQLite_JSGet},
    {"colCount",  2,   JSPROP_ENUMERATE|JSPROP_READONLY|JSPROP_PERMANENT,SQLite_JSGet},
    {"count",  1,   JSPROP_ENUMERATE|JSPROP_READONLY|JSPROP_PERMANENT,SQLite_JSGet},
    {"length",  1,   JSPROP_ENUMERATE|JSPROP_READONLY|JSPROP_PERMANENT,SQLite_JSGet},
    {"className",3, JSPROP_ENUMERATE|JSPROP_READONLY|JSPROP_PERMANENT,SQLite_JSGet},
    {"error",4,JSPROP_ENUMERATE|JSPROP_READONLY|JSPROP_PERMANENT,SQLite_JSGet},
    {"driverName",5,JSPROP_ENUMERATE|JSPROP_READONLY|JSPROP_PERMANENT,SQLite_JSGet},
    {0}
};

static JSFunctionSpec SQLite_functions[] = {
    {"data",     SQLite_GetDataC,      2},
    {"get",     SQLite_GetDataC,      2},
    {"set",     SQLite_SetDataC,      3},
    {"setN",     SQLite_SetDataD,      3},
    {"save",    SQLite_Save,          0},
    {"type",    SQLite_FieldType,         1},
    {"column",  SQLite_FindColumn,    1},
//    {"error",  SQLite_GetLastError,    0},
    {"title",  SQLite_ColumnTitle,    1},
    {"setTitle",  SQLite_SetTitle,    2},
    {"width",  SQLite_GetColWidth,    1},
    {"deleteColumn",  SQLite_DelCol,    1},
    {"addColumn",  SQLite_AddCol,    1},
    {"add",SQLite_AddRow ,1},
    {"addRow",SQLite_AddRow ,1},
    {"del",SQLite_DelRow,1},
    {"deleteRow",SQLite_DelRow,1},
    {"delRow",SQLite_DelRow,1},
    {"find",SQLite_FindRow,1},
    {"index",SQLite_Index,10},
    {"findNext",SQLite_FindNext,1},
    {"getRow",SQLite_GetRow,1},
    {"setRow",SQLite_SetRow,2},
    {"toString",SQLite_ToString,0},
    {"close",SQLite_Close,0},
    {"getMessage",SQLite_GetMessage,1},
    {"getHTML",SQLite_GetMessage2,1},
    {"append",SQLite_Append,3},
    {0}
};

static JSFunctionSpec SQLite_fnstatic[] = {
    {"create",SQLite_Create,2},
    {"help",  SQLite_HELP,    0},
    {0}
};

static JSClass SQLite_class = {
    "SQLite", JSCLASS_HAS_PRIVATE,         //SQLite_JSGet
    JS_PropertyStub,  JS_PropertyStub, JS_PropertyStub,   JS_PropertyStub,
    JS_EnumerateStub, JS_ResolveStub, JS_ConvertStub,SQLite_JSFinalize
};

JSObject*
SQLite_Object(JSContext *cx, SQLite* t,bool autodelete,JSPointerBase* Parent)
{
 JSObject* obj;
 GETENV;
 ENTERNATIVE(cx);
 MAKENEW(SQLite);
 SETPRIVATE(obj,SQLite,t,autodelete,Parent);

 return obj;
}

void SQLite_InitClass(JSContext *cx, JSObject *obj)
{
 GETENV;
 INITCLASS(SQLite);
}

JSClass* SQLite_Class() {return &SQLite_class;}

